import copy
import json
import typer
import os.path

class MalformedMaxMSP(Exception):
    def __init__(self, max_path, e):
        self.path = max_path
        self.err = e
    def __str__(self):
        return '[Error] Malformed MaxMSP file: %s\n\t%s' % (self.path, self.err.__str__())

class MaxParser:
    '''
    This class parses max msp files, which are stored as JSON.
    '''

    def __init__(self, max_f):
        '''
        Parses the json file and stores the patch as a class member
        Parameters:
            max_f: max patch file pointer
        '''
        try:
            self._path = max_f.name
        except AttributeError:
            pass

        try:
            self.maxmsp = json.load(max_f, strict=False)
        except (UnicodeDecodeError, ValueError) as e:
            raise MalformedMaxMSP(self._path, e)

        # more error checking, enforce some structure of JSON
        try:
            self.maxmsp["patcher"]
            self.maxmsp["patcher"]["boxes"]
            self.maxmsp["patcher"]["lines"]
        except KeyError as e:
            raise MalformedMaxMSP(self._path, e)

    def get_canvas_dims(self):
        '''
        Retrieves the dimensions of the Max MSP canvas on which the objects are rendered
        [x, y]
        '''
        return self.maxmsp["patcher"]["rect"][2:4]

    def get_objects(self):
        '''
        Retrieves all of the objects from the max patch
        '''
        return self.maxmsp["patcher"]["boxes"]

    def get_connections(self):
        '''
        Retrieves all of the patch cords from the max patch
        '''
        return self.maxmsp["patcher"]["lines"]

    def get_comments(self):
        '''
        Retrieves the max objects that are comments
        '''
        objects = self.get_objects()
        return [o for o in objects if o["box"]["maxclass"] == "comment"]

    def get_patcher(self):
        '''
        Retrieves the patcher
        '''
        return self.maxmsp["patcher"]

    def get_subpatches(self, objects):
        '''
        Retrieves the max objects and patch cords from max subpatches
        '''
        nested_objects = []
        nested_connections = []
        more_nested_patches = False

        objects = copy.deepcopy(objects) # prevent overwrite in renaming

        for o in objects:
            if 'patcher' in o["box"].keys():
                # print(o["box"])
                # print('Found patcher')
                more_nested_patches = True

                # subpatch IDs are independent of parent patch
                renamed_id_boxes = []
                for box in o['box']['patcher']["boxes"]:
                    box['box']['id'] = f"{o['box']['id']}_{box['box']['id']}"
                    renamed_id_boxes.append(box)

                nested_objects += renamed_id_boxes

                renamed_id_connections = []
                for l in o["box"]['patcher']["lines"]:
                    l['patchline']['destination'][0] = f"{o['box']['id']}_{l['patchline']['destination'][0]}"
                    l['patchline']['source'][0] = f"{o['box']['id']}_{l['patchline']['source'][0]}"

                    renamed_id_connections.append(l)

                nested_connections += renamed_id_connections

        if more_nested_patches:
            more_nested_objects, more_nested_connections = self.get_subpatches(nested_objects)

            nested_objects += more_nested_objects
            nested_connections += more_nested_connections

        return nested_objects, nested_connections

def strip_subpatch(patcher, box_keys, c_keys):
    patcher = copy.deepcopy(patcher)

    # delete patcher keys
    for k in copy.deepcopy(list(patcher.keys())):
        if k not in ['boxes', 'lines']:
            del patcher[k]

    for line in patcher["lines"]:
        for l in copy.deepcopy(list(line['patchline'].keys())):
            if l not in c_keys:
                del line['patchline'][l]

    for o in patcher["boxes"]:
        # delete main box level keys
        for k in copy.deepcopy(list(o['box'].keys())):
            if k not in box_keys + ['id', 'patcher']:
                del o['box'][k]

        if 'patcher' in o["box"].keys():

            for k in copy.deepcopy(list(o['box']['patcher'].keys())):
                if k not in ['boxes', 'lines']:
                    del o['box']['patcher'][k]

            for box in o['box']['patcher']["boxes"]:
                #  del keys inside patcher boxes
                for k in copy.deepcopy(list(box['box'].keys())):
                    if k not in box_keys + ['id', 'patcher']:
                        del box['box'][k]

                    if k == 'patcher':
                        o['box']['patcher'] = copy.deepcopy(strip_subpatch(o['box']['patcher'], box_keys, c_keys))
                        
            for line in  o['box']['patcher']["lines"]:
                for l in copy.deepcopy(list(line['patchline'].keys())):
                    if l not in c_keys:
                        del line['patchline'][l]

    return patcher

def generate_intermediate(patcher, box_keys, c_keys):

    intermediate_format = dict()

    for o in patcher["boxes"]:
        o_type = list(o.keys())[0]
        contents = copy.deepcopy(o[o_type])
        
        id = copy.deepcopy(contents['id'])
        del contents['id']

        for k in copy.deepcopy(list(contents.keys())):
            if k == 'patcher':
                contents[k] = generate_intermediate(strip_subpatch(contents[k], box_keys, c_keys), box_keys, c_keys)

            elif k not in box_keys:
                del contents[k]

        contents['type'] = o_type

        if id not in intermediate_format.keys():
            intermediate_format[id] = dict()
            intermediate_format[id]['serialized_contents'] = contents
            intermediate_format[id]['connections'] = []

        else:
            raise ValueError(f"{id} has already been created before")

    for c in patcher["lines"]:
        c_type = list(c.keys())[0]
        contents = copy.deepcopy(c[c_type])
        contents['type'] = c_type

        for k in copy.deepcopy(list(contents.keys())):
            if k not in c_keys:
                del contents[k]

        intermediate_format[contents['source'][0]]['connections'].append(contents)

    return intermediate_format

def main(filename: str):
    if os.path.isfile(filename):
        with open(filename) as f:
            pp = MaxParser(f)

            intermediate_format = dict()

            # keys to keep for object
            box_keys = ['maxclass', 'numinlets', 'numoutlets', 'text', 'outlettype', "minimum", "maximum", 'saved_attribute_attributes' ,'varname', 'save', 'items']
            # do not keep 'orientation', specifies horizontal or vertical does not affect functions
            # do not keep 'hint', it is just text that appears when you hover over patch
            # do not include 'comment', just a comment about the patch
            # 'saved_attribute_attributes' is default values for a patch

            # keys to keep for patchline
            c_keys = ['source', 'destination'] 
            # note that we do not include 'midpoints', because they are related to how patch lines can be placed on the editor
            # we do not include 'disabled' since it just prevents the pathcord from moving by locking it and does not affect anything else
            # we do not include 'hidden' since it just hides the patchcord, a patchcord can only be hidden when it's locked

            return generate_intermediate(pp.get_patcher(), box_keys, c_keys)
            

            # All the objects retrieved
            # import pprint
            # pprint.pprint({
            #     'objects': pp.get_objects(),
            #     'connections': pp.get_connections(),
            #     'canvas_dims': pp.get_canvas_dims()
            # })
    else:
        raise ValueError('File does not exist')

if __name__ == "__main__":
    typer.run(main)