import typer
import os.path
import copy

class MalformedPureData(Exception):
    def __init__(self, pd_path, e):
        self.path = pd_path
        self.err = e
    def __str__(self):
        return "[Error] Malformed Pure Data file: %s\n\t%s" % (self.path, self.err.__str__())

class PDParser:
    '''
    This class parses pure data files.
    '''

    def __init__(self, pd_f):
        '''
        Parses the pd save file and stores the patch as a class member
        '''

        try:
            self._path = pd_f.name
        except AttributeError:
            pass

        self.objects = []
        self.connections = []

        obj_convert = {
            "obj": "newobj",
            "msg": "message",
            "text": "comment"
        }

        if not hasattr(pd_f, 'readline') and hasattr(pd_f, 'read') and callable(pd_f.read):
            pd_f = pd_f.read().split('\n')
 
        try:
            patch_level = 0
            patch_level_name = 'PD-ROOT'
            previous_patch_level_name_stack = []
            line = ''
            for l in pd_f:
                if len(line):
                    line += str(" %s" % l).rstrip(' ')
                else:
                    line += str(l).strip(' ')

                if line.endswith(';\n') or line.endswith(';'):
                    # we now have all of the word-wrapped line
                    line = line.rstrip('\n')
                    line = line.rstrip(';')
                    line = line.strip()
                    entity_attrs = line.split(' ')
                    chunk_type = entity_attrs[0]
                    entity_type = entity_attrs[1]

                    xdim = ydim = 0

                    if chunk_type == '#N':
                        if entity_type == 'canvas':
                            patch_level += 1

                            if patch_level > 1:
                                obj_entity = {'box': {
                                    "id": f"{'_'.join(previous_patch_level_name_stack + [patch_level_name])}_" + "obj-%d" % self.get_subpatch_max_id(previous_patch_level_name_stack + [patch_level_name]),
                                    "maxclass": 'subpatch',
                                    "fontname": 'Arial',
                                    "numinlets": 0,
                                    "numoutlets": 0,
                                    "text": entity_attrs[6:],
                                    'original_text': entity_attrs
                                }}
                                self.objects.append(obj_entity)

                            if len(entity_attrs) == 8:
                                previous_patch_level_name_stack.append(patch_level_name)
                                patch_level_name = entity_attrs[6]

                        if patch_level == 1:
                            # parse the global patch level canvas element
                            self.xoffset = int(entity_attrs[2])
                            self.yoffset = int(entity_attrs[3])
                            self.xdim = int(entity_attrs[4])
                            self.ydim = int(entity_attrs[5])
                            try:
                                self.fontsize = int(entity_attrs[6]) 
                            except IndexError:
                                self.fontsize = 10
                    elif chunk_type == '#X' and entity_type in ['restore', 'pop']:
                        if entity_type == 'restore' and len(entity_attrs) == 5:
                            # https://puredata.info/docs/developer/PdFileFormat#r39
                            obj_entity = {'box': {
                                "id": f"{'_'.join(previous_patch_level_name_stack + [patch_level_name])}_" + "obj-%d" % self.get_subpatch_max_id(previous_patch_level_name_stack + [patch_level_name]),
                                "maxclass": 'newobj',
                                "fontname": 'Arial',
                                "fontsize": 10,
                                "text": entity_attrs[4:],
                                "numinlets": 0,
                                "numoutlets": 0,
                                "patching_rect": [int(entity_attrs[2]), int(entity_attrs[3]), xdim, ydim],
                                'original_text': entity_attrs
                            }}
                            self.objects.append(obj_entity)

                        patch_level_name = previous_patch_level_name_stack.pop()
                        patch_level -= 1
                    elif chunk_type == '#X':
                        if entity_type == "connect":
                            conn_entity = {"patchline": {
                                "destination": [f"{'_'.join(previous_patch_level_name_stack + [patch_level_name])}_" + "obj-%s" % entity_attrs[4], int(entity_attrs[5])],
                                "source": [f"{'_'.join(previous_patch_level_name_stack + [patch_level_name])}_" + "obj-%s" % entity_attrs[2], int(entity_attrs[3])]
                            }}
                            self.connections.append(conn_entity)
                        else:
                            obj_entity = {'box': {
                                "id": f"{'_'.join(previous_patch_level_name_stack + [patch_level_name])}_" + "obj-%d" % self.get_subpatch_max_id(previous_patch_level_name_stack + [patch_level_name]),
                                "numinlets": 0,
                                "numoutlets": 0,
                                "text": "",
                                "fontsize": 'Arial',
                                "fontsize": 10,
                                "patching_rect": [int(entity_attrs[2]), int(entity_attrs[3]), xdim, ydim] if entity_type!= 'declare' else []
                            }}

                            # some objects have preset dimensions, like bangs or toggles
                            if entity_type == "floatatom":
                                obj_entity["box"]["text"] = '0'     # does not store number in file format
                            elif entity_type == "msg" or entity_type == "text":
                                try:
                                    obj_entity["box"]["text"] = ' '.join(entity_attrs[4:])
                                except IndexError:
                                    pass
                            elif entity_type == "obj":
                                obj_type = entity_attrs[4]
                                if obj_type == "bng" or obj_type == "tgl":
                                    xdim = entity_attrs[5]
                                    ydim = entity_attrs[5]
                                elif obj_type == "nbx":
                                    obj_entity["box"]["text"] = '0'
                                    ydim = entity_attrs[6]
                                    xdim = int(entity_attrs[5]) * int(entity_attrs[17]) # numdigits * digitsize in pixels
                                elif obj_type == "vsl" or obj_type == "hsl" or obj_type == "vu":
                                    xdim = entity_attrs[5]
                                    ydim = entity_attrs[6]
                                elif obj_type == "vradio":
                                    ydim = entity_attrs[5]
                                    xdim = entity_attrs[15]
                                elif obj_type == "hradio":
                                    xdim = entity_attrs[5]
                                    ydim = entity_attrs[15]
                                elif obj_type == "outlet":
                                    xdim = entity_attrs[2]
                                    ydim = entity_attrs[3]
                                    obj_entity["box"]["text"] = ' '.join(entity_attrs[4:])
                                else:
                                    # it's some other built in object
                                    obj_entity["box"]["text"] = ' '.join(entity_attrs[4:])

                            # update calculated object dimensions ... some still need to be done on the clientside
                            # depending on the drawing surface dpi
                            obj_entity["box"]["patching_rect"][2:] = [int(xdim), int(ydim)]

                            # sometimes there are empty message boxes or comments
                            # better to ask forgiveness
                            try:
                                obj_entity["box"]["maxclass"] = obj_convert[entity_type]
                            except KeyError:
                                obj_entity["box"]["maxclass"] = entity_type

                            self.objects.append(obj_entity)
                    line = ''

            # calculate implicit properties of objects that are stated explicitly in max but not pd
            for c in self.connections:
                source_obj_id = c['patchline']['source'][0]
                target_obj_id = c['patchline']['destination'][0]

                matched_source = False
                matched_target = False

                for obj in self.objects:
                    # print(obj['box']['id'])
                    if source_obj_id == obj['box']['id']:
                        obj['box']['numoutlets'] += 1
                        matched_source = True

                    if target_obj_id == obj['box']['id']:
                        obj['box']['numinlets'] += 1
                        matched_target = True

                if not matched_source or not matched_target:
                    raise IndexError(f'Unmatched {source_obj_id} {matched_source} {target_obj_id} {matched_target}')

        except (KeyError, ValueError, IndexError) as e:
            raise MalformedPureData(self._path, e)

    def get_subpatch_max_id(self, hierarchy):
        id = 0
        string_map = f"{'_'.join(hierarchy)}_obj-"
        for obj in self.objects:
            if obj['box']['id'].startswith(string_map):
                id+=1
        return id

    def get_canvas_dims(self):
        '''
        Retrieves the dimensions of the canvas on which the objects are rendered
        [x, y]
        '''
        return (self.xdim, self.ydim) 

    def get_objects(self):
        '''
        Retrieves all of the objects from the pd patch
        '''
        return self.objects

    def get_connections(self):
        '''
        Retrieves all of the patch cords from the pd patch
        '''
        return self.connections

    def get_comments(self):
        '''
        Retrieves the pd objects that are comments
        '''
        return [o for o in self.objects if o['box']['maxclass'] == "comment"]

def main(filename: str):
    if os.path.isfile(filename):
        with open(filename) as f:
            pp = PDParser(f)

            intermediate_format = dict()

            keys = ['maxclass', 'numinlets', 'numoutlets', 'text']

            # import pprint
            # pprint.pprint(pp.get_objects())

            for o in pp.get_objects():
                o_type = list(o.keys())[0]
                contents = copy.deepcopy(o[o_type])
                
                id = copy.deepcopy(contents['id'])
                del contents['id']

                for k in copy.deepcopy(list(contents.keys())):
                    if k not in keys:
                        del contents[k]

                contents['type'] = o_type

                if id not in intermediate_format.keys():
                    intermediate_format[id] = dict()
                    intermediate_format[id]['serialized_contents'] = contents
                    intermediate_format[id]['connections'] = []

                else:
                    raise ValueError(f"{id} has already been created before")

            for c in pp.get_connections():
                c_type = list(c.keys())[0]
                contents = copy.deepcopy(c[c_type])
                contents['type'] = c_type

                intermediate_format[contents['source'][0]]['connections'].append(contents)

            return intermediate_format

            # import pprint
            # # print(intermediate_format)
            # pprint.pprint(intermediate_format)

            # All the objects retrieved
            # import pprint
#             pprint.pprint({
#                 'objects': pp.get_objects(),
#                 'connections': pp.get_connections(),
#                 'canvas_dims': pp.get_canvas_dims()
#             })
    else:
        raise ValueError('File does not exist')

if __name__ == "__main__":
    typer.run(main)
