import subprocess
from deepdiff import DeepDiff, Delta, extract
import pprint
import re
import os
import tempfile
import pickle
from pydriller import Repository
import copy


import sys
original_path = sys.path
sys.path.append("..") 
import importlib
maxparser = importlib.import_module('Visual-Code-SZZ.parsers.max.maxparser')
pdparser = importlib.import_module('Visual-Code-SZZ.parsers.pure_data.pdparser')
sys.path = original_path
import logging


modified_keys = {
    "type_changes",
    "values_changed",
    # "unprocessed",
    # "iterable_item_moved",
    "repetition_change",
}

# These are keys we want to skip
# They are new nodes added to the IR
added_keys = {
    # "set_item_added",
    # "attribute_added",
    "dictionary_item_added",
    "iterable_item_added"
}

removed_keys = {
    "dictionary_item_removed",
    "iterable_item_removed"
    # "set_item_removed",
    # "attribute_removed"
}

def flatten(l):
    return [item for sublist in l for item in sublist]

def get_short_sha(long_sha):
    return subprocess.check_output(['git', 'rev-parse', '--short', long_sha]).decode('utf-8').strip()

def check_extension(path, parser_type):
    if path is None:
        return False
        
    ext = os.path.splitext(path)[-1]
    if parser_type == 'pd' and ext == '.pd':
        return True
    elif parser_type == 'max' and ext == '.maxpat':
        return True
    elif parser_type == 'max' and ext == '.maxhelp':
        return True
    return False

def szz_vc(fixing_sha, main_branch, gr, parser_type, debug=False, cache_key=""):
    gr.checkout(fixing_sha)
    fix_commit = gr.get_commit(fixing_sha)
    fixing_sha = fix_commit.hash

    # Get all the valid files of a fix commit
    filepaths = [f.new_path for f in fix_commit.modified_files if f is not None and check_extension(f.new_path, parser_type)]

    cache = {}

    diff_lookup = {}

    one_changedepth_inducing_commits_per_file = {}
    max_changedepth_inducing_commits_per_file = {}

    for f in filepaths:

        # Get all the commits for the modified file
        commits = gr.get_commits_modified_file(f)

        # Go to next file if there is no commits (nothing to blame)
        if len(commits) == 0:
            continue

        if debug:
            print(f'Total number of commits that modified (including renaming) {f}: {len(commits)}')

        commits_before_fix_count = 0

        # Trace through commits to get filenames and commits to skip because they can't be retrieved
        error_commits_to_skip = set()
        # Get valid file names
        all_filenames = set()
        all_filenames.add(f)
        for c in commits:
            prev_commit = gr.get_commit(c)

            # Check if the commits are in the main branch
            if main_branch not in prev_commit.branches:
                continue

            # Skip merge commits
            if prev_commit.merge:
                continue
            
            try:
                for mod_file in prev_commit.modified_files:
                    if mod_file.new_path in all_filenames or mod_file.old_path in all_filenames:
                        if mod_file.new_path is not None:
                            all_filenames.add(mod_file.new_path)
                        if mod_file.old_path is not None:
                            all_filenames.add(mod_file.old_path)
            except ValueError:
                error_commits_to_skip.add(prev_commit.hash)
                # Could not retrieve commit
                pass

        # Build intermediate formats for file
        prev_intermediate_format = {}
        for c in commits[::-1]:
            # Iterate from oldest to newest commits since we need to build diffs from oldest to newest
            prev_commit = gr.get_commit(c)

            if prev_commit.hash in error_commits_to_skip:
                continue

            # Skip merge commits
            if prev_commit.merge:
                continue

            # Check if the commits are in the main branch
            if main_branch not in prev_commit.branches:
                continue

            for mod_file in prev_commit.modified_files:
                # If the file hasn't been seen, then skip
                if mod_file.new_path not in all_filenames and mod_file.old_path not in all_filenames:
                    continue

                # If source code is empty, then skip
                if mod_file.source_code is None:
                    continue

                # Parse the code into intermediate representation
                with tempfile.NamedTemporaryFile(delete=False) as fp:
                    fp.write(mod_file.source_code.encode())

                if parser_type == 'max':
                    intermediate_format = maxparser.main(fp.name)
                elif parser_type == 'pd':
                    intermediate_format = pdparser.main(fp.name)

                os.remove(fp.name)

                # Perform diff of the intermediate representation
                diff = DeepDiff(prev_intermediate_format, intermediate_format, ignore_order=True, report_repetition=True)

                prev_intermediate_format = intermediate_format

                if f not in diff_lookup.keys():
                    diff_lookup[f] = {}

                diff_lookup[f][c] = {
                    'diff': diff,
                    'affected_nodes': diff.affected_root_keys,
                    'affected_paths': diff.affected_paths,
                    'modified_paths': flatten([v for k,v in diff.to_dict().items() if k in modified_keys]),
                    'added_paths': flatten([v for k,v in diff.to_dict().items() if k in added_keys]),
                    'removed_paths': flatten([v for k,v in diff.to_dict().items() if k in removed_keys]),
                    'commit_date': prev_commit.committer_date,
                    'intermediate_format': intermediate_format
                }

                root_regex = r"(root\['[a-zA-Z-_]*_?obj-\d+'\])"
                node_regex= r"(root\['[a-zA-Z-_]*_?obj-\d+'\])\[*"

                # Figure out when nodes were added from, deleted from, or modified from
                diff_lookup[f][c]['modified_nodes'] = list(set([re.search(node_regex, elem).group(1) for elem in diff_lookup[f][c]['modified_paths'] if re.search(node_regex, elem)]))
                diff_lookup[f][c]['added_nodes'] = list(set([re.search(node_regex, elem).group(1) for elem in diff_lookup[f][c]['added_paths'] if re.search(node_regex, elem)]))
                diff_lookup[f][c]['deleted_nodes'] = list(set([re.search(node_regex, elem).group(1) for elem in diff_lookup[f][c]['removed_paths'] if re.search(node_regex, elem)]))

                # Figure out when nodes were added, deleted, or modified
                # diff_lookup[f][c]['deleted_nodes'] = [elem for elem in diff_lookup[f][c]['removed_paths'] if elem.count('[') == 1] # no longer exist in the commit
                # diff_lookup[f][c]['added_nodes'] = [elem for elem in diff_lookup[f][c]['added_paths'] if elem.count('[') == 1] # new in the commit
                # diff_lookup[f][c]['modified_nodes'] = list(set([re.search(root_regex, elem).group(1) for elem in diff_lookup[f][c]['modified_paths'] if re.search(root_regex, elem)]))

                if fixing_sha not in diff_lookup[f].keys():
                    commits_before_fix_count+=1

                if debug:
                    print(get_short_sha(c), mod_file.old_path, mod_file.new_path, prev_commit.committer_date, len(intermediate_format.keys()), f'| Number of modified files {len(prev_commit.modified_files)}', f'| Nodes-touched: {len(diff.affected_root_keys)}')

                break

        # Run Visual Code SZZ for file
        commits_last_modified_nodes = set()
        commits_last_modified_nodes_paths = set()

        if debug:
            print("\n\n1 Change Depth SZZ")
            print("added nodes")
            for p in diff_lookup[f][fixing_sha]['added_nodes']:
                print(p)

            print('\n\n')

        commits_list_oldest_newest = list(diff_lookup[f].keys())
        ind = commits_list_oldest_newest.index(fixing_sha)
        sha_before_fixing_sha = commits_list_oldest_newest[ind-1]

        commits_list_newest_oldest = commits_list_oldest_newest[::-1]
        commits_to_traverse_from = commits_list_newest_oldest[commits_list_newest_oldest.index(sha_before_fixing_sha)+1:]

        # 1 Change depth implementation
        for path in diff_lookup[f][fixing_sha]['deleted_nodes'] + diff_lookup[f][fixing_sha]['modified_nodes'] + diff_lookup[f][fixing_sha]['added_nodes']:
            fixing_value = ''
            if path not in diff_lookup[f][fixing_sha]['deleted_nodes']:
                try:
                    fixing_value = str(extract(diff_lookup[f][fixing_sha]['intermediate_format'], path))
                except (KeyError, IndexError):
                    pass

            if path in diff_lookup[f][sha_before_fixing_sha]['added_nodes']:
                # Node was added in previous commit and removed or modified in fixing commit
                if debug:
                    path_value = str(extract(diff_lookup[f][sha_before_fixing_sha]['intermediate_format'], path))
                    print(f"{sha_before_fixing_sha} | node {path} |\n value before inducing: <added in the commit before fixing commit> |\n inducing value: {path_value} |\n fixing_value: {fixing_value} \n\n")

                commits_last_modified_nodes.add(sha_before_fixing_sha)
                continue

            try:
                path_value = str(extract(diff_lookup[f][sha_before_fixing_sha]['intermediate_format'], path))
            except (KeyError, IndexError) as e:
                if path in diff_lookup[f][fixing_sha]['added_nodes']:
                    # Special case where node hasn't existed before, so just move on
                    continue
                else:
                    raise Exception(e)

            for index, c in enumerate(commits_to_traverse_from):
                v = diff_lookup[f][c]

                if path in v['added_nodes']:
                    # This is the first commit when it was introduced
                    if debug:
                        print(f"{c} | node {path} |\n value before inducing: <added in commit before inducing commit> |\n inducing value: {path_value} |\n fixing_value: {fixing_value} \n\n")
                    commits_last_modified_nodes.add(c)
                    break
 
                try:
                    prev_path_value =  str(extract(v['intermediate_format'], path))
                except (KeyError, IndexError):
                    # Probably from missing a commit
                    if debug:
                        print(f"{commits_to_traverse_from[index-1]} | node {path} |\n value before inducing: {prev_path_value} |\n inducing value: {path_value} |\n fixing_value: {fixing_value} \n\n")
                    commits_last_modified_nodes.add(commits_to_traverse_from[index-1])
                    break

                if prev_path_value != path_value:
                    if debug:
                        print(f"{c} | node {path} |\n value before inducing: {prev_path_value} |\n inducing value: {path_value} |\n fixing_value: {fixing_value} \n\n")
                    commits_last_modified_nodes.add(c)
                    break

                elif prev_path_value == path_value and (index == len(commits_to_traverse_from) - 1):
                    # case when prev_path_value remains the same throughout all commits
                    if debug:
                        print(f"{c} | node {path} |\n value before inducing: <beginning of commits> |\n inducing value: {path_value} |\n fixing_value: {fixing_value} \n\n")
                    commits_last_modified_nodes.add(c)
                    break

        if debug:
            print("\n\n\nMax Change Depth SZZ-VC")
            print("added paths")
            for p in diff_lookup[f][fixing_sha]['added_paths']:
                print(p)

            print('\n\n')

        for path in diff_lookup[f][fixing_sha]['added_paths']:
            # Special case where we try to find prior change of this path
            ######
            # Try to go a level up until we get a value extracted, then we try the most recent change to that level
            one_level_up = copy.deepcopy(path)
            path_value = '!@#$%^&*()-'
            while one_level_up.count(']') > 1 and path_value == '!@#$%^&*()-':
                try:
                    path_value = str(extract(diff_lookup[f][sha_before_fixing_sha]['intermediate_format'], one_level_up))
                except (KeyError, IndexError) as e:
                    last_bracket_idx = one_level_up.rfind('[')
                    one_level_up = one_level_up[:last_bracket_idx]

            if path_value == '!@#$%^&*()-':
                continue
            ######
            # Return one_level_up which is the new path
            ######

            fixing_value = ''
            try:
                fixing_value = str(extract(diff_lookup[f][fixing_sha]['intermediate_format'], one_level_up))
            except (KeyError, IndexError):
                pass

            if one_level_up in diff_lookup[f][sha_before_fixing_sha]['added_paths']:
                # Node was added in previous commit and removed or modified in fixing commit
                if debug:
                    path_value = str(extract(diff_lookup[f][sha_before_fixing_sha]['intermediate_format'], path))
                    print(f"{sha_before_fixing_sha} | {one_level_up} of {path} |\n value before inducing: <added in the commit before fixing commit> |\n inducing value: {path_value} |\n fixing_value: {fixing_value} \n\n")

                commits_last_modified_nodes_paths.add(sha_before_fixing_sha)
                continue

            for index, c in enumerate(commits_to_traverse_from):
                v = diff_lookup[f][c]

                if one_level_up in v['added_paths']:
                    # This is the first commit when it was introduced
                    if debug:
                        print(f"{c} | {one_level_up} of {path} |\n value before inducing: <added in commit before inducing commit> |\n inducing value: {path_value} |\n fixing_value: {fixing_value} \n\n")
                    commits_last_modified_nodes_paths.add(c)
                    break
 
                try:
                    prev_path_value =  str(extract(v['intermediate_format'], one_level_up))
                except (KeyError, IndexError):
                    # Probably from missing a commit
                    if debug:
                        print(f"{commits_to_traverse_from[index-1]} | {one_level_up} of {path} |\n value before inducing: {prev_path_value} |\n inducing value: {path_value} |\n fixing_value: {fixing_value} \n\n")
                    commits_last_modified_nodes_paths.add(commits_to_traverse_from[index-1])
                    break

                if prev_path_value != path_value:
                    if debug:
                        print(f"{c} | {one_level_up} of {path} |\n value before inducing: {prev_path_value} |\n inducing value: {path_value} |\n fixing_value: {fixing_value} \n\n")
                    commits_last_modified_nodes_paths.add(c)
                    break

                elif prev_path_value == path_value and (index == len(commits_to_traverse_from) - 1):
                    # case when prev_path_value remains the same throughout all commits
                    if debug:
                        print(f"{c} | {one_level_up} of {path} |\n value before inducing: <beginning of commits> |\n inducing value: {path_value} |\n fixing_value: {fixing_value} \n\n")
                    commits_last_modified_nodes_paths.add(c)
                    break

        # Max change depth implementation
        for path in diff_lookup[f][fixing_sha]['removed_paths'] + diff_lookup[f][fixing_sha]['modified_paths']:
            fixing_value = ''
            if path not in diff_lookup[f][fixing_sha]['removed_paths']:
                try:
                    fixing_value = str(extract(diff_lookup[f][fixing_sha]['intermediate_format'], path))
                except (KeyError, IndexError):
                    pass

            if path in diff_lookup[f][sha_before_fixing_sha]['added_paths']:
                # Node was added in previous commit and removed or modified in fixing commit
                if debug:
                    path_value = str(extract(diff_lookup[f][sha_before_fixing_sha]['intermediate_format'], path))
                    print(f"{sha_before_fixing_sha} | {path} |\n value before inducing: <added in the commit before fixing commit> |\n inducing value: {path_value} |\n fixing_value: {fixing_value} \n\n")

                commits_last_modified_nodes_paths.add(sha_before_fixing_sha)
                continue

            path_value = str(extract(diff_lookup[f][sha_before_fixing_sha]['intermediate_format'], path))

            # if path.count('[') == 1:
            #     # Special case the whole node has been removed
            #     # Look for when the node last touched
            #     print(path, 'node removed')

            for index, c in enumerate(commits_to_traverse_from):
                v = diff_lookup[f][c]

                if path in v['added_paths']:
                    # This is the first commit when it was introduced
                    if debug:
                        print(f"{c} | {path} |\n value before inducing: <added in commit before inducing commit> |\n inducing value: {path_value} |\n fixing_value: {fixing_value} \n\n")
                    commits_last_modified_nodes_paths.add(c)
                    break
 
                try:
                    prev_path_value =  str(extract(v['intermediate_format'], path))
                except (KeyError, IndexError):
                    # Probably from missing a commit
                    if debug:
                        print(f"{commits_to_traverse_from[index-1]} | {path} |\n value before inducing: {prev_path_value} |\n inducing value: {path_value} |\n fixing_value: {fixing_value} \n\n")
                    commits_last_modified_nodes_paths.add(commits_to_traverse_from[index-1])
                    break

                if prev_path_value != path_value:
                    if debug:
                        print(f"{c} | {path} |\n value before inducing: {prev_path_value} |\n inducing value: {path_value} |\n fixing_value: {fixing_value} \n\n")
                    commits_last_modified_nodes_paths.add(c)
                    break

                elif prev_path_value == path_value and (index == len(commits_to_traverse_from) - 1):
                     # case when prev_path_value remains the same throughout all commits
                    if debug:
                        print(f"{c} | {path} |\n value before inducing: <beginning of commits> |\n inducing value: {path_value} |\n fixing_value: {fixing_value} \n\n")
                    commits_last_modified_nodes_paths.add(c)
                    break

        for x in commits_last_modified_nodes:
            one_changedepth_inducing_commits_per_file.setdefault(f, set()).add(x)

        for x in commits_last_modified_nodes_paths:
            max_changedepth_inducing_commits_per_file.setdefault(f, set()).add(x)

    gr.reset()
    return {
        'one': one_changedepth_inducing_commits_per_file,
        'max': max_changedepth_inducing_commits_per_file,
    }
